@@ -1,6 +1,6 @@
 /*
    This file is part of GNUnet.
-   Copyright (C) 2021--2022 GNUnet e.V.
+   Copyright (C) 2021--2024 GNUnet e.V.
 
    GNUnet is free software: you can redistribute it and/or modify it
    under the terms of the GNU Affero General Public License as published
@@ -34,7 +34,6 @@
  */
 /**@{*/
 
-#include <gnunet/platform.h>
 #include <gnunet/gnunet_common.h>
 #include <gnunet/gnunet_time_lib.h>
 #include <gnunet/gnunet_util_lib.h>
@@ -44,7 +43,7 @@
  *                          the #GNUNET_MESSENGER_VERSION of the GNUnet Messenger
  *                          service while the patch version is independent.
  */
-#define GNUNET_CHAT_VERSION 0x000000010000L
+#define GNUNET_CHAT_VERSION 0x000000030000L
 
 #define GNUNET_CHAT_VERSION_MAJOR ((GNUNET_CHAT_VERSION >> 32L) & 0xFFFFL)
 #define GNUNET_CHAT_VERSION_MINOR ((GNUNET_CHAT_VERSION >> 16L) & 0xFFFFL)
@@ -72,51 +71,57 @@ enum GNUNET_CHAT_MessageKind
    */
   GNUNET_CHAT_KIND_LOGIN = 3,      /**< GNUNET_CHAT_KIND_LOGIN */
 
+  /**
+   * The kind to inform that the application needs to cleanup 
+   * resources related to the currently connected account.
+   */
+  GNUNET_CHAT_KIND_LOGOUT = 4,     /**< GNUNET_CHAT_KIND_LOGOUT */
+
   /**
    * The kind to inform that a context was updated.
    */
-  GNUNET_CHAT_KIND_UPDATE = 4,      /**< GNUNET_CHAT_KIND_UPDATE */
+  GNUNET_CHAT_KIND_UPDATE = 5,     /**< GNUNET_CHAT_KIND_UPDATE */
 
   /**
    * The kind to inform that a contact has joined a chat.
    */
-  GNUNET_CHAT_KIND_JOIN = 5,       /**< GNUNET_CHAT_KIND_JOIN */
+  GNUNET_CHAT_KIND_JOIN = 6,       /**< GNUNET_CHAT_KIND_JOIN */
 
   /**
    * The kind to inform that a contact has left a chat.
    */
-  GNUNET_CHAT_KIND_LEAVE = 6,      /**< GNUNET_CHAT_KIND_LEAVE */
+  GNUNET_CHAT_KIND_LEAVE = 7,      /**< GNUNET_CHAT_KIND_LEAVE */
 
   /**
    * The kind to inform that a contact has changed.
    */
-  GNUNET_CHAT_KIND_CONTACT = 7,    /**< GNUNET_CHAT_KIND_CONTACT */
+  GNUNET_CHAT_KIND_CONTACT = 8,    /**< GNUNET_CHAT_KIND_CONTACT */
 
   /**
    * The kind to describe an invitation to a different chat.
    */
-  GNUNET_CHAT_KIND_INVITATION = 8, /**< GNUNET_CHAT_KIND_INVITATION */
+  GNUNET_CHAT_KIND_INVITATION = 9, /**< GNUNET_CHAT_KIND_INVITATION */
 
   /**
    * The kind to describe a text message.
    */
-  GNUNET_CHAT_KIND_TEXT = 9,       /**< GNUNET_CHAT_KIND_TEXT */
+  GNUNET_CHAT_KIND_TEXT = 10,      /**< GNUNET_CHAT_KIND_TEXT */
 
   /**
    * The kind to describe a shared file.
    */
-  GNUNET_CHAT_KIND_FILE = 10,      /**< GNUNET_CHAT_KIND_FILE */
-
-  /**
-   * The kind to describe a whispered message.
-   */
-  GNUNET_CHAT_KIND_WHISPER = 11,   /**< GNUNET_CHAT_KIND_WHISPER */
+  GNUNET_CHAT_KIND_FILE = 11,      /**< GNUNET_CHAT_KIND_FILE */
 
   /**
    * The kind to inform about a deletion of a previous message.
    */
   GNUNET_CHAT_KIND_DELETION = 12,  /**< GNUNET_CHAT_KIND_DELETION */
 
+  /**
+   * The kind to tag a previous message.
+   */
+  GNUNET_CHAT_KIND_TAG = 13,       /**< GNUNET_CHAT_KIND_TAG */
+
   /**
    * An unknown kind of message.
    */
@@ -173,6 +178,11 @@ struct GNUNET_CHAT_File;
  */
 struct GNUNET_CHAT_Invitation;
 
+/**
+ * Struct of a chat ticket.
+ */
+struct GNUNET_CHAT_Ticket;
+
 /**
  * Iterator over chat accounts of a specific chat handle.
  *
@@ -181,10 +191,24 @@ struct GNUNET_CHAT_Invitation;
  * @param[in,out] account Chat account
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef int
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_AccountCallback) (void *cls,
-				const struct GNUNET_CHAT_Handle *handle,
-				struct GNUNET_CHAT_Account *account);
+                                const struct GNUNET_CHAT_Handle *handle,
+                                struct GNUNET_CHAT_Account *account);
+
+/**
+ * Method called for each attribute of a specific chat handle.
+ *
+ * @param[in,out] cls Closure from #GNUNET_CHAT_get_attributes
+ * @param[in] handle Chat handle
+ * @param[in] name Attribute name
+ * @param[in] value Attribute value
+ */
+typedef void
+(*GNUNET_CHAT_AttributeCallback) (void *cls,
+                                  struct GNUNET_CHAT_Handle *handle,
+                                  const char *name,
+                                  const char *value);
 
 /**
  * Method called when a lobby is opened to share with others via a chat URI.
@@ -193,8 +217,8 @@ typedef int
  * @param[in] uri Chat URI of the lobby or NULL on error
  */
 typedef void
-(*GNUNET_CHAT_LobbyCallback) (void *cls,
-			      const struct GNUNET_CHAT_Uri *uri);
+(*GNUNET_CHAT_LobbyCallback) (void *cls, 
+                              const struct GNUNET_CHAT_Uri *uri);
 
 /**
  * Iterator over chat contacts of a specific chat handle.
@@ -204,10 +228,23 @@ typedef void
  * @param[in,out] contact Chat contact
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef int
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_ContactCallback) (void *cls,
                                 struct GNUNET_CHAT_Handle *handle,
-				struct GNUNET_CHAT_Contact *contact);
+                                struct GNUNET_CHAT_Contact *contact);
+
+/**
+ * Iterator over chat tickets of a specific chat contact.
+ *
+ * @param[in,out] cls Closure
+ * @param[in] contact Chat contact
+ * @param[in,out] ticket Chat ticket
+ * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
+ */
+typedef enum GNUNET_GenericReturnValue
+(*GNUNET_CHAT_ContactTicketCallback) (void *cls,
+                                      const struct GNUNET_CHAT_Contact *contact,
+                                      struct GNUNET_CHAT_Ticket *ticket);
 
 /**
  * Iterator over chat groups of a specific chat handle.
@@ -217,10 +254,10 @@ typedef int
  * @param[in,out] group Chat group
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef int
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_GroupCallback) (void *cls,
                               struct GNUNET_CHAT_Handle *handle,
-			      struct GNUNET_CHAT_Group *group);
+                              struct GNUNET_CHAT_Group *group);
 
 /**
  * Iterator over chat contacts in a specific chat group.
@@ -230,7 +267,7 @@ typedef int
  * @param[in,out] contact Chat contact
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef int
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_GroupContactCallback) (void *cls,
                                      const struct GNUNET_CHAT_Group *group,
                                      struct GNUNET_CHAT_Contact *contact);
@@ -243,10 +280,10 @@ typedef int
  * @param[in] message Chat message
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef int
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_ContextMessageCallback) (void *cls,
                                        struct GNUNET_CHAT_Context *context,
-				       const struct GNUNET_CHAT_Message *message);
+                                       const struct GNUNET_CHAT_Message *message);
 
 /**
  * Iterator over chat files in a specific chat context.
@@ -256,10 +293,10 @@ typedef int
  * @param[in,out] file Chat file
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef int
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_ContextFileCallback) (void *cls,
                                     struct GNUNET_CHAT_Context *context,
-				    struct GNUNET_CHAT_File *file);
+                                    struct GNUNET_CHAT_File *file);
 
 /**
  * Iterator over chat contacts in a chat to check whether they received a
@@ -272,11 +309,22 @@ typedef int
  * 			   #GNUNET_NO otherwise
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef int
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_MessageReadReceiptCallback) (void *cls,
                                            const struct GNUNET_CHAT_Message *message,
-					   const struct GNUNET_CHAT_Contact *contact,
-					   int read_receipt);
+                                           const struct GNUNET_CHAT_Contact *contact,
+                                           int read_receipt);
+
+/**
+ * Iterator over chat messages.
+ *
+ * @param[in,out] cls Closure from #GNUNET_CHAT_context_iterate_messages
+ * @param[in] message Chat message
+ * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
+ */
+typedef enum GNUNET_GenericReturnValue
+(*GNUNET_CHAT_MessageCallback) (void *cls,
+                                const struct GNUNET_CHAT_Message *message);
 
 /**
  * Method called during an upload of a specific file in a chat to share it.
@@ -289,8 +337,8 @@ typedef int
 typedef void
 (*GNUNET_CHAT_FileUploadCallback) (void *cls,
                                    const struct GNUNET_CHAT_File *file,
-				   uint64_t completed,
-				   uint64_t size);
+                                   uint64_t completed,
+                                   uint64_t size);
 
 /**
  * Method called during a download of a specific file in a chat which was shared.
@@ -303,8 +351,8 @@ typedef void
 typedef void
 (*GNUNET_CHAT_FileDownloadCallback) (void *cls,
                                      const struct GNUNET_CHAT_File *file,
-				     uint64_t completed,
-				     uint64_t size);
+                                     uint64_t completed,
+                                     uint64_t size);
 
 /**
  * Method called during an unindexing of a specific file in a chat which was
@@ -318,8 +366,22 @@ typedef void
 typedef void
 (*GNUNET_CHAT_FileUnindexCallback) (void *cls,
                                     struct GNUNET_CHAT_File *file,
-				    uint64_t completed,
-				    uint64_t size);
+                                    uint64_t completed,
+                                    uint64_t size);
+
+/**
+ * Method called during a chat ticket consumption for each of its attributes.
+ *
+ * @param[in,out] cls Closure from #GNUNET_CHAT_ticket_consume
+ * @param[in] ticket Chat ticket
+ * @param[in] name Attribute name
+ * @param[in] value Attribute value
+ */
+typedef void
+(*GNUNET_CHAT_TicketAttributeCallback) (void *cls,
+                                        const struct GNUNET_CHAT_Ticket *ticket,
+                                        const char *name,
+                                        const char *value);
 
 /**
  * Start a chat handle with a certain configuration.
@@ -334,8 +396,8 @@ typedef void
  */
 struct GNUNET_CHAT_Handle*
 GNUNET_CHAT_start (const struct GNUNET_CONFIGURATION_Handle *cfg,
-		   GNUNET_CHAT_ContextMessageCallback msg_cb,
-		   void *msg_cls);
+                   GNUNET_CHAT_ContextMessageCallback msg_cb,
+                   void *msg_cls);
 
 /**
  * Stops a chat handle closing all its remaining resources and frees the
@@ -357,9 +419,9 @@ GNUNET_CHAT_stop (struct GNUNET_CHAT_Handle *handle);
  * @param[in] name Account name
  * @return #GNUNET_OK on success, #GNUNET_NO on failure and otherwise #GNUNET_SYSERR
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_account_create (struct GNUNET_CHAT_Handle *handle,
-			    const char* name);
+                            const char* name);
 
 /**
  * Deletes an existing chat account of a given chat <i>handle</i> under a
@@ -369,9 +431,9 @@ GNUNET_CHAT_account_create (struct GNUNET_CHAT_Handle *handle,
  * @param[in] name Account name
  * @return #GNUNET_OK on success, otherwise #GNUNET_SYSERR
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_account_delete(struct GNUNET_CHAT_Handle *handle,
-			   const char* name);
+                           const char* name);
 
 /**
  * Iterates through the accounts of a given chat <i>handle</i> with a selected
@@ -384,8 +446,8 @@ GNUNET_CHAT_account_delete(struct GNUNET_CHAT_Handle *handle,
  */
 int
 GNUNET_CHAT_iterate_accounts (const struct GNUNET_CHAT_Handle *handle,
-			      GNUNET_CHAT_AccountCallback callback,
-			      void *cls);
+                              GNUNET_CHAT_AccountCallback callback,
+                              void *cls);
 
 /**
  * Connects a chat <i>handle</i> to a selected chat <i>account</i>.
@@ -394,7 +456,7 @@ GNUNET_CHAT_iterate_accounts (const struct GNUNET_CHAT_Handle *handle,
  */
 void
 GNUNET_CHAT_connect (struct GNUNET_CHAT_Handle *handle,
-		     const struct GNUNET_CHAT_Account *account);
+                     const struct GNUNET_CHAT_Account *account);
 
 /**
  * Disconnects a chat <i>handle</i> from the current chat account.
@@ -424,7 +486,7 @@ GNUNET_CHAT_get_connected (const struct GNUNET_CHAT_Handle *handle);
  * @param[in,out] handle Chat handle
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_update (struct GNUNET_CHAT_Handle *handle);
 
 /**
@@ -434,9 +496,9 @@ GNUNET_CHAT_update (struct GNUNET_CHAT_Handle *handle);
  * @param[in] name New name or NULL
  * @return #GNUNET_YES on success, #GNUNET_NO on failure and #GNUNET_SYSERR if <i>handle</i> is NULL
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_set_name (struct GNUNET_CHAT_Handle *handle,
-		      const char *name);
+                      const char *name);
 
 /**
  * Returns the name of a chat handle for related communication or NULL if no
@@ -457,6 +519,45 @@ GNUNET_CHAT_get_name (const struct GNUNET_CHAT_Handle *handle);
 const char*
 GNUNET_CHAT_get_key (const struct GNUNET_CHAT_Handle *handle);
 
+/**
+ * Updates an attribute of a chat handle for related communication under a given 
+ * <i>name</i> and a custom <i>value</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] name Attribute name
+ * @param[in] value Attribute value
+ * @param[in] expires Expiration time or NULL
+ */
+void
+GNUNET_CHAT_set_attribute (struct GNUNET_CHAT_Handle *handle,
+                           const char *name,
+                           const char *value,
+                           const struct GNUNET_TIME_Relative *expires);
+
+/**
+ * Deletes an attribute of a chat handle for related communication under a given
+ * <i>name</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] name Attribute name
+ */
+void
+GNUNET_CHAT_delete_attribute (struct GNUNET_CHAT_Handle *handle,
+                              const char *name);
+
+/**
+ * Calls an optional <i>callback</i> for each attribute of a given chat 
+ * <i>handle</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] callback Callback for attribute iteration (optional)
+ * @param[in,out] cls Closure for attribute iteration (optional)
+ */
+void
+GNUNET_CHAT_get_attributes (struct GNUNET_CHAT_Handle *handle,
+                            GNUNET_CHAT_AttributeCallback callback,
+                            void *cls);
+
 /**
  * Convert an UTF-8 String to a chat URI which will be newly allocated.
  *
@@ -466,7 +567,7 @@ GNUNET_CHAT_get_key (const struct GNUNET_CHAT_Handle *handle);
  */
 struct GNUNET_CHAT_Uri*
 GNUNET_CHAT_uri_parse (const char *uri,
-		       char **emsg);
+                       char **emsg);
 
 /**
  * Convert a chat URI to a UTF-8 String.
@@ -496,9 +597,9 @@ GNUNET_CHAT_uri_destroy (struct GNUNET_CHAT_Uri *uri);
  */
 struct GNUNET_CHAT_Lobby*
 GNUNET_CHAT_lobby_open (struct GNUNET_CHAT_Handle *handle,
-			struct GNUNET_TIME_Relative delay,
-			GNUNET_CHAT_LobbyCallback callback,
-			void *cls);
+                        struct GNUNET_TIME_Relative delay,
+                        GNUNET_CHAT_LobbyCallback callback,
+                        void *cls);
 
 /**
  * Closes a chat <i>lobby</i> overriding the expiration to be now.
@@ -517,7 +618,7 @@ GNUNET_CHAT_lobby_close (struct GNUNET_CHAT_Lobby *lobby);
  */
 void
 GNUNET_CHAT_lobby_join (struct GNUNET_CHAT_Handle *handle,
-			const struct GNUNET_CHAT_Uri *uri);
+                        const struct GNUNET_CHAT_Uri *uri);
 
 
 /**
@@ -529,7 +630,7 @@ GNUNET_CHAT_lobby_join (struct GNUNET_CHAT_Handle *handle,
  */
 void
 GNUNET_CHAT_set_user_pointer (struct GNUNET_CHAT_Handle *handle,
-			      void *user_pointer);
+                              void *user_pointer);
 
 /**
  * Returns the custom user pointer of a given chat <i>handle</i> or NULL if it
@@ -552,8 +653,8 @@ GNUNET_CHAT_get_user_pointer (const struct GNUNET_CHAT_Handle *handle);
  */
 int
 GNUNET_CHAT_iterate_contacts (struct GNUNET_CHAT_Handle *handle,
-			      GNUNET_CHAT_ContactCallback callback,
-			      void *cls);
+                              GNUNET_CHAT_ContactCallback callback,
+                              void *cls);
 
 /**
  * Returns the provided name of a given <i>account</i> or NULL on failure.
@@ -573,7 +674,7 @@ GNUNET_CHAT_account_get_name (const struct GNUNET_CHAT_Account *account);
  */
 void
 GNUNET_CHAT_account_set_user_pointer (struct GNUNET_CHAT_Account *account,
-				      void *user_pointer);
+                                      void *user_pointer);
 
 /**
  * Returns the custom user pointer of a given <i>contact</i> or NULL if it was
@@ -600,7 +701,7 @@ GNUNET_CHAT_account_get_user_pointer (const struct GNUNET_CHAT_Account *account)
  */
 struct GNUNET_CHAT_Group*
 GNUNET_CHAT_group_create (struct GNUNET_CHAT_Handle *handle,
-			  const char* topic);
+                          const char* topic);
 
 /**
  * Iterates through the groups of a given chat <i>handle</i> with a selected
@@ -613,8 +714,8 @@ GNUNET_CHAT_group_create (struct GNUNET_CHAT_Handle *handle,
  */
 int
 GNUNET_CHAT_iterate_groups (struct GNUNET_CHAT_Handle *handle,
-			    GNUNET_CHAT_GroupCallback callback,
-			    void *cls);
+                            GNUNET_CHAT_GroupCallback callback,
+                            void *cls);
 
 /**
  * Leaves the private chat with a specific <i>contact</i> and frees the
@@ -623,7 +724,7 @@ GNUNET_CHAT_iterate_groups (struct GNUNET_CHAT_Handle *handle,
  * @param[in,out] contact Cntact
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_contact_delete (struct GNUNET_CHAT_Contact *contact);
 
 /**
@@ -635,7 +736,7 @@ GNUNET_CHAT_contact_delete (struct GNUNET_CHAT_Contact *contact);
  */
 void
 GNUNET_CHAT_contact_set_name (struct GNUNET_CHAT_Contact *contact,
-			      const char *name);
+                              const char *name);
 
 /**
  * Returns the provided name of a given <i>contact</i> or its custom nick name
@@ -675,7 +776,7 @@ GNUNET_CHAT_contact_get_context (struct GNUNET_CHAT_Contact *contact);
  */
 void
 GNUNET_CHAT_contact_set_user_pointer (struct GNUNET_CHAT_Contact *contact,
-				      void *user_pointer);
+                                      void *user_pointer);
 
 /**
  * Returns the custom user pointer of a given <i>contact</i> or NULL if it was
@@ -692,11 +793,80 @@ GNUNET_CHAT_contact_get_user_pointer (const struct GNUNET_CHAT_Contact *contact)
  * it has sent messages with.
  *
  * @param[in] contact Contact
- * @return GNUNET_YES if the contact is owned, otherwise GNUNET_NO
- *         and GNUNET_SYSERR on failure
+ * @return #GNUNET_YES if the contact is owned, otherwise# GNUNET_NO
+ *         and #GNUNET_SYSERR on failure
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_contact_is_owned (const struct GNUNET_CHAT_Contact *contact);
+
+/**
+ * Blocks or unblocks a given <i>contact</i> by the current account.
+ *
+ * @param[in,out] contact Contact
+ * @param[in] blocked #GNUNET_YES to block, #GNUNET_NO to unblock
+ */
+void
+GNUNET_CHAT_contact_set_blocked (struct GNUNET_CHAT_Contact *contact,
+                                 enum GNUNET_GenericReturnValue blocked);
+
+/**
+ * Returns if a given <i>contact</i> is blocked by the current account and
+ * whether messages of it should be filtered.
+ *
+ * @param[in] contact Contact
+ * @return #GNUNET_YES if the contact is blocked, #GNUNET_SYSERR on failure and 
+ *         #GNUNET_NO otherwise
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_contact_is_blocked (const struct GNUNET_CHAT_Contact *contact);
+
+/**
+ * Tags a given <i>contact</i> by the current account with a specific
+ * <i>tag</i>.
+ *
+ * @param[in,out] contact Contact
+ * @param[in] tag Tag
+ */
+void
+GNUNET_CHAT_contact_tag (struct GNUNET_CHAT_Contact *contact,
+                         const char *tag);
+
+/**
+ * Untags a given <i>contact</i> by the current account with a specific
+ * <i>tag</i>.
+ *
+ * @param[in,out] contact Contact
+ * @param[in] tag Tag
+ */
+void
+GNUNET_CHAT_contact_untag (struct GNUNET_CHAT_Contact *contact,
+                           const char *tag);
+
+/**
+ * Returns if a given <i>contact</i> is tagged by the current account with
+ * a specific <i>tag</i>.
+ *
+ * @param[in] contact Contact
+ * @return #GNUNET_YES if the contact is tagged, #GNUNET_SYSERR on failure and 
+ *         #GNUNET_NO otherwise
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_contact_is_tagged (const struct GNUNET_CHAT_Contact *contact,
+                               const char *tag);
+
+/**
+ * Iterates through the tickets of a given <i>contact</i> with a selected
+ * callback and custom closure.
+ *
+ * @param[in,out] contact Contact
+ * @param[in] callback Callback for ticket iteration (optional)
+ * @param[in,out] cls Closure for ticket iteration (optional)
+ * @return Amount of tickets iterated
  */
 int
-GNUNET_CHAT_contact_is_owned (const struct GNUNET_CHAT_Contact *contact);
+GNUNET_CHAT_contact_iterate_tickets (const struct GNUNET_CHAT_Contact *contact,
+                                     GNUNET_CHAT_ContactTicketCallback callback,
+                                     void *cls);
 
 /**
  * Leaves a specific <i>group</i> chat and frees its memory if it is not shared
@@ -705,7 +875,7 @@ GNUNET_CHAT_contact_is_owned (const struct GNUNET_CHAT_Contact *contact);
  * @param[in,out] group Group
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_group_leave (struct GNUNET_CHAT_Group *group);
 
 /**
@@ -717,7 +887,7 @@ GNUNET_CHAT_group_leave (struct GNUNET_CHAT_Group *group);
  */
 void
 GNUNET_CHAT_group_set_name (struct GNUNET_CHAT_Group *group,
-			    const char *name);
+                            const char *name);
 
 /**
  * Returns the custom nick name of a given <i>group</i> if it was overriden.
@@ -737,7 +907,7 @@ GNUNET_CHAT_group_get_name (const struct GNUNET_CHAT_Group *group);
  */
 void
 GNUNET_CHAT_group_set_user_pointer (struct GNUNET_CHAT_Group *group,
-				    void *user_pointer);
+                                    void *user_pointer);
 
 /**
  * Returns the custom user pointer of a given <i>group</i> or NULL if it was
@@ -758,7 +928,7 @@ GNUNET_CHAT_group_get_user_pointer (const struct GNUNET_CHAT_Group *group);
  */
 void
 GNUNET_CHAT_group_invite_contact (const struct GNUNET_CHAT_Group *group,
-				  struct GNUNET_CHAT_Contact *contact);
+                                  struct GNUNET_CHAT_Contact *contact);
 
 /**
  * Iterates through the contacts of a given <i>group</i> with a selected
@@ -771,8 +941,8 @@ GNUNET_CHAT_group_invite_contact (const struct GNUNET_CHAT_Group *group,
  */
 int
 GNUNET_CHAT_group_iterate_contacts (const struct GNUNET_CHAT_Group *group,
-				    GNUNET_CHAT_GroupContactCallback callback,
-				    void *cls);
+                                    GNUNET_CHAT_GroupContactCallback callback,
+                                    void *cls);
 
 /**
  * Sets a custom <i>user pointer</i> to a given <i>member</i> relative to a
@@ -784,8 +954,8 @@ GNUNET_CHAT_group_iterate_contacts (const struct GNUNET_CHAT_Group *group,
  */
 void
 GNUNET_CHAT_member_set_user_pointer (struct GNUNET_CHAT_Group *group,
-				     const struct GNUNET_CHAT_Contact *member,
-				     void *user_pointer);
+                                     const struct GNUNET_CHAT_Contact *member,
+                                     void *user_pointer);
 
 /**
  * Returns the custom user pointer of a given <i>member</i> relative to a
@@ -797,7 +967,7 @@ GNUNET_CHAT_member_set_user_pointer (struct GNUNET_CHAT_Group *group,
  */
 void*
 GNUNET_CHAT_member_get_user_pointer (const struct GNUNET_CHAT_Group *group,
-				     const struct GNUNET_CHAT_Contact *member);
+                                     const struct GNUNET_CHAT_Contact *member);
 
 /**
  * Returns the chat context for a chat with a given <i>group</i>.
@@ -816,7 +986,7 @@ GNUNET_CHAT_group_get_context (struct GNUNET_CHAT_Group *group);
  * @return #GNUNET_OK if usable, #GNUNET_NO if the context has been requested,
  *         #GNUNET_SYSERR otherwise.
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_context_get_status (const struct GNUNET_CHAT_Context *context);
 
 /**
@@ -857,7 +1027,7 @@ GNUNET_CHAT_context_get_group (struct GNUNET_CHAT_Context *context);
  */
 void
 GNUNET_CHAT_context_set_user_pointer (struct GNUNET_CHAT_Context *context,
-				      void *user_pointer);
+                                      void *user_pointer);
 
 /**
  * Returns the custom user pointer of a given chat <i>context</i> or NULL if it
@@ -876,9 +1046,9 @@ GNUNET_CHAT_context_get_user_pointer (const struct GNUNET_CHAT_Context *context)
  * @param[in] text Text
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_context_send_text (struct GNUNET_CHAT_Context *context,
-			       const char *text);
+                               const char *text);
 
 /**
  * Sends a read receipt depending on a selected <i>message</i> into a given
@@ -888,9 +1058,9 @@ GNUNET_CHAT_context_send_text (struct GNUNET_CHAT_Context *context,
  * @param[in] message Message (optional)
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_context_send_read_receipt (struct GNUNET_CHAT_Context *context,
-				       const struct GNUNET_CHAT_Message *message);
+                                       const struct GNUNET_CHAT_Message *message);
 
 /**
  * Uploads a local file specified via its <i>path</i> using symmetric encryption
@@ -905,9 +1075,9 @@ GNUNET_CHAT_context_send_read_receipt (struct GNUNET_CHAT_Context *context,
  */
 struct GNUNET_CHAT_File*
 GNUNET_CHAT_context_send_file (struct GNUNET_CHAT_Context *context,
-			       const char *path,
-			       GNUNET_CHAT_FileUploadCallback callback,
-			       void *cls);
+                               const char *path,
+                               GNUNET_CHAT_FileUploadCallback callback,
+                               void *cls);
 
 /**
  * Shares the information to download and decrypt a specific <i>file</i> from
@@ -917,9 +1087,23 @@ GNUNET_CHAT_context_send_file (struct GNUNET_CHAT_Context *context,
  * @param[in] file File handle
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_context_share_file (struct GNUNET_CHAT_Context *context,
-				const struct GNUNET_CHAT_File *file);
+                                const struct GNUNET_CHAT_File *file);
+
+/**
+ * Sends a tag message targeting a selected <i>message</i> into a given
+ * chat <i>context</i> with a given <i>tag</i> value.
+ *
+ * @param[in,out] context Chat context
+ * @param[in] message Message
+ * @param[in] tag Tag value
+ * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_context_send_tag (struct GNUNET_CHAT_Context *context,
+                              const struct GNUNET_CHAT_Message *message,
+                              const char *tag);
 
 /**
  * Iterates through the contacts of a given chat <i>context</i> with a selected
@@ -932,8 +1116,8 @@ GNUNET_CHAT_context_share_file (struct GNUNET_CHAT_Context *context,
  */
 int
 GNUNET_CHAT_context_iterate_messages (struct GNUNET_CHAT_Context *context,
-				      GNUNET_CHAT_ContextMessageCallback callback,
-				      void *cls);
+                                      GNUNET_CHAT_ContextMessageCallback callback,
+                                      void *cls);
 
 /**
  * Iterates through the files of a given chat <i>context</i> with a selected
@@ -946,8 +1130,8 @@ GNUNET_CHAT_context_iterate_messages (struct GNUNET_CHAT_Context *context,
  */
 int
 GNUNET_CHAT_context_iterate_files (struct GNUNET_CHAT_Context *context,
-				   GNUNET_CHAT_ContextFileCallback callback,
-				   void *cls);
+                                   GNUNET_CHAT_ContextFileCallback callback,
+                                   void *cls);
 
 /**
  * Returns the kind of a given <i>message</i> to determine its content and
@@ -978,26 +1162,81 @@ struct GNUNET_CHAT_Contact*
 GNUNET_CHAT_message_get_sender (const struct GNUNET_CHAT_Message *message);
 
 /**
- * Returns #GNUNET_YES if the message was sent by the related chat handle,
- * otherwise it returns #GNUNET_NO.
+ * Returns the contact of the recipient from a given <i>message</i>.
+ *
+ * @param[in] message Message
+ * @return Contact of the messages recipient or NULL
+ */
+struct GNUNET_CHAT_Contact*
+GNUNET_CHAT_message_get_recipient (const struct GNUNET_CHAT_Message *message);
+
+/**
+ * Returns #GNUNET_YES if the <i>message</i> was sent by the related chat 
+ * handle, otherwise it returns #GNUNET_NO.
  *
  * @param[in] message Message
  * @return #GNUNET_YES if the message was sent, otherwise #GNUNET_NO
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_message_is_sent (const struct GNUNET_CHAT_Message *message);
 
 /**
- * Returns #GNUNET_YES if the message was received privately encrypted by the
- * related chat handle, otherwise it returns #GNUNET_NO.
+ * Returns #GNUNET_YES if the <i>message</i> was received privately encrypted
+ * by the related chat handle, otherwise it returns #GNUNET_NO.
  *
  * @param[in] message Message
  * @return #GNUNET_YES if the message was privately received,
  * 	   otherwise #GNUNET_NO
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_message_is_private (const struct GNUNET_CHAT_Message *message);
 
+/**
+ * Returns #GNUNET_YES if the <i>message</i> was received recently by related 
+ * chat handle, otherwise it returns #GNUNET_NO.
+ *
+ * @param[in] message Message
+ * @return #GNUNET_YES if the message was received recently,
+ *     otherwise #GNUNET_NO 
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_message_is_recent (const struct GNUNET_CHAT_Message *message);
+
+/**
+ * Returns #GNUNET_YES if the <i>message</i> was received because of an 
+ * update by related chat handle, otherwise it returns #GNUNET_NO.
+ *
+ * @param[in] message Message
+ * @return #GNUNET_YES if the message was received to update
+ *     a previous message, otherwise #GNUNET_NO 
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_message_is_update (const struct GNUNET_CHAT_Message *message);
+
+/**
+ * Returns #GNUNET_YES if the <i>message</i> was received because of a 
+ * deletion by related chat handle or if it has been deleted internally, 
+ * otherwise it returns #GNUNET_NO.
+ *
+ * @param[in] message Message
+ * @return #GNUNET_YES if the message was received to delete
+ *     a previous message, otherwise #GNUNET_NO 
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_message_is_deleted (const struct GNUNET_CHAT_Message *message);
+
+/**
+ * Returns #GNUNET_YES if the <i>message</i> was tagged with a custom
+ * <i>tag</i> value, otherwise it returns #GNUNET_NO.
+ *
+ * @param[in] message Message
+ * @return #GNUNET_YES if the message was tagged with
+ *     the given tag value, otherwise #GNUNET_NO
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_message_is_tagged (const struct GNUNET_CHAT_Message *message,
+                               const char *tag);
+
 /**
  * Iterates through the contacts of the context related to a given chat
  * <i>message</i> to check whether it was received by each of the contacts.
@@ -1009,8 +1248,8 @@ GNUNET_CHAT_message_is_private (const struct GNUNET_CHAT_Message *message);
  */
 int
 GNUNET_CHAT_message_get_read_receipt (const struct GNUNET_CHAT_Message *message,
-				      GNUNET_CHAT_MessageReadReceiptCallback callback,
-				      void *cls);
+                                      GNUNET_CHAT_MessageReadReceiptCallback callback,
+                                      void *cls);
 
 /**
  * Returns the text of a given <i>message</i> if its kind is
@@ -1062,9 +1301,23 @@ GNUNET_CHAT_message_get_target (const struct GNUNET_CHAT_Message *message);
  * @param[in] delay Relative delay
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_message_delete (const struct GNUNET_CHAT_Message *message,
-			    struct GNUNET_TIME_Relative delay);
+                            struct GNUNET_TIME_Relative delay);
+
+/**
+ * Iterates through the tag messages in the context of a given 
+ * <i>message</i>.
+ *
+ * @param[in] message Message
+ * @param[in] callback Callback for tag message iteration (optional)
+ * @param[in,out] cls Closure for tag message iteration (optional)
+ * @return Amount of tag messages iterated or #GNUNET_SYSERR on failure
+ */
+int
+GNUNET_CHAT_message_iterate_tags (const struct GNUNET_CHAT_Message *message,
+                                  GNUNET_CHAT_MessageCallback callback,
+                                  void *cls);
 
 /**
  * Returns the name of a given <i>file</i> handle.
@@ -1111,7 +1364,7 @@ GNUNET_CHAT_file_get_local_size (const struct GNUNET_CHAT_File *file);
  * @param[in] file File handle
  * @return #GNUNET_YES during active upload, #GNUNET_NO otherwise
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_is_uploading (const struct GNUNET_CHAT_File *file);
 
 /**
@@ -1148,7 +1401,7 @@ GNUNET_CHAT_file_close_preview (struct GNUNET_CHAT_File *file);
  */
 void
 GNUNET_CHAT_file_set_user_pointer (struct GNUNET_CHAT_File *file,
-				   void *user_pointer);
+                                   void *user_pointer);
 
 /**
  * Returns the custom user pointer of a given <i>file</i> handle or NULL if it
@@ -1166,7 +1419,7 @@ GNUNET_CHAT_file_get_user_pointer (const struct GNUNET_CHAT_File *file);
  * @param[in] file File handle
  * @return #GNUNET_YES during active download, #GNUNET_NO otherwise
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_is_downloading (const struct GNUNET_CHAT_File *file);
 
 /**
@@ -1178,10 +1431,10 @@ GNUNET_CHAT_file_is_downloading (const struct GNUNET_CHAT_File *file);
  * @param[in,out] cls Closure for file downloading (optional)
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_start_download (struct GNUNET_CHAT_File *file,
-				 GNUNET_CHAT_FileDownloadCallback callback,
-				 void *cls);
+                                 GNUNET_CHAT_FileDownloadCallback callback,
+                                 void *cls);
 
 /**
  * Pause downloading a file from a given <i>file</i> handle.
@@ -1189,7 +1442,7 @@ GNUNET_CHAT_file_start_download (struct GNUNET_CHAT_File *file,
  * @param[in,out] file File handle
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_pause_download (struct GNUNET_CHAT_File *file);
 
 /**
@@ -1198,7 +1451,7 @@ GNUNET_CHAT_file_pause_download (struct GNUNET_CHAT_File *file);
  * @param[in,out] file File handle
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_resume_download (struct GNUNET_CHAT_File *file);
 
 /**
@@ -1207,7 +1460,7 @@ GNUNET_CHAT_file_resume_download (struct GNUNET_CHAT_File *file);
  * @param[in,out] file File handle
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_stop_download (struct GNUNET_CHAT_File *file);
 
 /**
@@ -1216,7 +1469,7 @@ GNUNET_CHAT_file_stop_download (struct GNUNET_CHAT_File *file);
  * @param[in] file File handle
  * @return #GNUNET_YES during active unindexing, #GNUNET_NO otherwise
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_is_unindexing (const struct GNUNET_CHAT_File *file);
 
 /**
@@ -1228,10 +1481,10 @@ GNUNET_CHAT_file_is_unindexing (const struct GNUNET_CHAT_File *file);
  * @param[in,out] cls Closure for file unindexing (optional)
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_unindex (struct GNUNET_CHAT_File *file,
-			  GNUNET_CHAT_FileUnindexCallback callback,
-			  void *cls);
+                          GNUNET_CHAT_FileUnindexCallback callback,
+                          void *cls);
 
 /**
  * Accepts a given chat <i>invitation</i> to enter another chat.
@@ -1241,6 +1494,54 @@ GNUNET_CHAT_file_unindex (struct GNUNET_CHAT_File *file,
 void
 GNUNET_CHAT_invitation_accept (struct GNUNET_CHAT_Invitation *invitation);
 
+/**
+ * Rejects a given chat <i>invitation</i> to enter another chat.
+ *
+ * @param[in,out] invitation Chat invitation
+ */
+void
+GNUNET_CHAT_invitation_reject (struct GNUNET_CHAT_Invitation *invitation);
+
+/**
+ * Returns if a given <i>invitation</i> got accepted.
+ *
+ * @param[in] invitation Chat invitation
+ * @return #GNUNET_YES if accepted, #GNUNET_NO otherwise
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_invitation_is_accepted (const struct GNUNET_CHAT_Invitation *invitation);
+
+/**
+ * Returns if a given <i>invitation</i> got rejected.
+ *
+ * @param[in] invitation Chat invitation
+ * @return #GNUNET_YES if rejected, #GNUNET_NO otherwise
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_invitation_is_rejected (const struct GNUNET_CHAT_Invitation *invitation);
+
+/**
+ * Returns the contact of the issuer from a given chat <i>ticket</i>.
+ *
+ * @param[in] invitation Chat invitation
+ * @return Chat contact
+ */
+const struct GNUNET_CHAT_Contact*
+GNUNET_CHAT_ticket_get_contact (const struct GNUNET_CHAT_Ticket *ticket);
+
+/**
+ * Consumes a given chat <i>ticket</i> and calls an optional <i>callback</i>
+ * for each of its attributes.
+ *
+ * @param[in,out] ticket Chat ticket
+ * @param[in] callback Callback for ticket consumption (optional)
+ * @param[in,out] cls Closure for ticket consumption (optional)
+ */
+void
+GNUNET_CHAT_ticket_consume (struct GNUNET_CHAT_Ticket *ticket,
+                            GNUNET_CHAT_TicketAttributeCallback callback,
+                            void *cls);
+
 /**@}*/
 
 #endif /* GNUNET_CHAT_LIB_H_ */
